package dev.fitko.fitconnect.core.crypto;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWEHeader;
import com.nimbusds.jose.JWEObject;
import com.nimbusds.jose.Payload;
import com.nimbusds.jose.crypto.RSADecrypter;
import com.nimbusds.jose.crypto.RSAEncrypter;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.exceptions.internal.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.internal.EncryptionException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;

import java.security.NoSuchAlgorithmException;
import java.text.ParseException;

import static dev.fitko.fitconnect.core.crypto.CryptoConstants.DEFAULT_JWE_ALGORITHM;
import static dev.fitko.fitconnect.core.crypto.CryptoConstants.DEFAULT_JWE_COMPRESSION_ALGORITHM;
import static dev.fitko.fitconnect.core.crypto.CryptoConstants.DEFAULT_JWE_ENCRYPTION_METHOD;
import static dev.fitko.fitconnect.core.crypto.CryptoConstants.DEFAULT_SYMMETRIC_ENCRYPTION_ALGORITHM;

public class JWECryptoService implements CryptoService {

    private static final Logger LOGGER = LoggerFactory.getLogger(JWECryptoService.class);
    private static final ObjectMapper MAPPER = new ObjectMapper();

    private final MessageDigestService messageDigestService;

    public JWECryptoService(final MessageDigestService messageDigestService) {
        this.messageDigestService = messageDigestService;
    }

    @Override
    public byte[] decryptToBytes(final RSAKey privateKey, final String encryptedData) throws DecryptionException {
        return decrypt(privateKey, encryptedData).toBytes();
    }

    @Override
    public String encryptObject(final RSAKey encryptionKey, final Object obj, final String contentType) throws EncryptionException {
        try {
            final Payload payload = new Payload(MAPPER.writeValueAsString(obj));
            return encrypt(encryptionKey, payload, contentType);
        } catch (final JsonProcessingException e) {
            throw new EncryptionException(e.getMessage(), e);
        }
    }

    @Override
    public String encryptBytes(final RSAKey publicKey, final byte[] bytes, final String contentType) throws EncryptionException {
        final Payload payload = new Payload(bytes);
        return encrypt(publicKey, payload, contentType);
    }

    @Override
    public String hashBytes(final byte[] data) {
        final byte[] hash = messageDigestService.createHash(data);
        return messageDigestService.toHexString(hash);
    }

    private String encrypt(final RSAKey publicKey, final Payload payload, final String contentType) throws EncryptionException {
        try {
            final KeyGenerator keyGenerator = KeyGenerator.getInstance(DEFAULT_SYMMETRIC_ENCRYPTION_ALGORITHM);
            keyGenerator.init(DEFAULT_JWE_ENCRYPTION_METHOD.cekBitLength());
            final SecretKey cek = keyGenerator.generateKey();
            return encryptPayload(publicKey, payload, cek, contentType);
        } catch (final NoSuchAlgorithmException | JOSEException e) {
            throw new EncryptionException(e.getMessage(), e);
        }
    }

    private Payload decrypt(final RSAKey privateKey, final String encData) throws DecryptionException {
        try {
            final var start = StopWatch.start();
            printAvailableVmMemory();
            final JWEObject jwe = JWEObject.parse(encData);
            jwe.decrypt(new RSADecrypter(privateKey));
            LOGGER.info("Decrypting {} bytes took {} ", encData.getBytes().length, StopWatch.stopWithFormattedTime(start));
            return jwe.getPayload();
        } catch (final ParseException | IllegalStateException | JOSEException e) {
            throw new DecryptionException(e.getMessage(), e);
        }
    }

    private JWEHeader getJWEHeader(final String keyID, final String contentType) {
        return new JWEHeader.Builder(DEFAULT_JWE_ALGORITHM, DEFAULT_JWE_ENCRYPTION_METHOD)
                .compressionAlgorithm(DEFAULT_JWE_COMPRESSION_ALGORITHM)
                .contentType(contentType)
                .keyID(keyID)
                .build();
    }

    private RSAEncrypter getEncrypter(final RSAKey publicKey, final SecretKey cek) {
        try {
            return new RSAEncrypter(publicKey.toRSAPublicKey(), cek);
        } catch (final JOSEException e) {
            throw new EncryptionException("RSAEncrypter could not be initialized", e);
        }
    }

    private String encryptPayload(final RSAKey publicKey, final Payload payload, final SecretKey cek, final String contentType) throws JOSEException, EncryptionException {
        final String keyID = getIdFromPublicKey(publicKey);
        final JWEObject jwe = new JWEObject(getJWEHeader(keyID, contentType), payload);
        final var start = StopWatch.start();
        jwe.encrypt(getEncrypter(publicKey, cek));
        printAvailableVmMemory();
        checkIfJWEObjectIsEncrypted(jwe);
        final String serializedEncryptedPayload = jwe.serialize();
        LOGGER.info("Encrypting {} bytes took {} ", payload.toBytes().length, StopWatch.stopWithFormattedTime(start));
        return serializedEncryptedPayload;
    }


    private void checkIfJWEObjectIsEncrypted(final JWEObject jwe) throws EncryptionException {
        if (!jwe.getState().equals(JWEObject.State.ENCRYPTED)) {
            throw new EncryptionException("JWE object is not encrypted");
        }
    }

    private String getIdFromPublicKey(final RSAKey publicKey) throws EncryptionException {
        final String keyID = publicKey.getKeyID();
        if (keyID == null || keyID.isEmpty()) {
            throw new EncryptionException("public key has no keyID");
        }
        return keyID;
    }

    private void printAvailableVmMemory() {
        LOGGER.info("Currently used VM memory {} MB", (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / 1024 / 1024);
    }
}
